;;; -*- Mode: Lisp; Package: EDITOR -*-
;;;
;;; This file is not part of xyzzy.
;;;

(provide "giraffe-mode")

(in-package "editor")

(export '(*giraffe-mode-hook* *giraffe-indent-tabs-mode* *giraffe-mode-map* *giraffe-mode-syntax-table* *giraffe-indent-level*))
(export '(giraffe-mode giraffe-indent-line giraffe-newline-and-indent giraffe-comment-region giraffe-uncomment-region))

(defvar *giraffe-indent-tabs-mode* nil)

(defvar *giraffe-mode-hook* nil)

(defvar *giraffe-indent-level* 2)

(defvar *giraffe-keyword-hash-table* nil)
(defvar *giraffe-keyword-file* "giraffe")


(defvar *giraffe-mode-syntax-table* nil)
(setq *giraffe-mode-syntax-table* nil)
(unless *giraffe-mode-syntax-table*
	(setq *giraffe-mode-syntax-table* (make-syntax-table))
	(set-syntax-string *giraffe-mode-syntax-table* #\")
	(set-syntax-string *giraffe-mode-syntax-table* #\')
	(set-syntax-escape *giraffe-mode-syntax-table* #\\)
	(set-syntax-symbol *giraffe-mode-syntax-table* #\_)
	(set-syntax-symbol *giraffe-mode-syntax-table* #\-)
	(set-syntax-match *giraffe-mode-syntax-table* #\( #\))
	(set-syntax-match *giraffe-mode-syntax-table* #\[ #\])
	(set-syntax-match *giraffe-mode-syntax-table* #\{ #\})
	(set-syntax-start-multi-comment *giraffe-mode-syntax-table* "/*")
	(set-syntax-end-multi-comment *giraffe-mode-syntax-table* "*/")
	(set-syntax-start-c++-comment *giraffe-mode-syntax-table* #\/)
	(set-syntax-end-c++-comment *giraffe-mode-syntax-table* #\LFD)
	; (set-syntax-start-comment *giraffe-mode-syntax-table* #\#)
	;	(set-syntax-end-comment *giraffe-mode-syntax-table* #\LFD nil t)
	)

(defvar *giraffe-mode-map* nil)
(unless *giraffe-mode-map*
	(setq *giraffe-mode-map* (make-sparse-keymap))
	(define-key *giraffe-mode-map* #\RET 'giraffe-newline-and-indent)
	(define-key *giraffe-mode-map* #\TAB 'giraffe-indent-line)
	(define-key *giraffe-mode-map* #\] 'giraffe-end-block)
	(define-key *giraffe-mode-map* #\) 'giraffe-end-block)
	(define-key *giraffe-mode-map* #\} 'giraffe-end-block)
	(define-key *giraffe-mode-map* #\C-h 'backward-delete-char-untabify-or-selection))


;------begin indent functions, referred ruby.l---------------------
; c-indent-line
(defun giraffe-indent-line ()
	(interactive "*")
	(if (or (not (interactive-p))
					*c-tab-always-indent*
					(save-excursion
						(skip-chars-backward " \t")
						(bolp)))
			; ς
			(let ((column (calc-giraffe-indent)))
				(when (integerp column)
					(save-excursion
						(goto-bol)
						(delete-region (point)
													 (progn
														 (skip-chars-forward " \t")
														 (point)))
						(indent-to column)))
				(if (and (bolp) column)
						(skip-chars-forward " \t")))
		(insert "\t"))
	t)

(defun giraffe-empty-line-p ()
	(save-excursion
		(goto-bol)
		(looking-at "\\([ \t]\\|/\\*\\([^*]\\|\\*[^/]\\)*\\*/\\)*\\($\\|#\\|//\\)")))

(defun giraffe-previous-line ()
	(while (backward-line)
		;(message "~D" (current-line-number))
		(unless (giraffe-empty-line-p) (return-from giraffe-previous-line t)))
	nil)

;What this searchs for is the indent of the line which contains the most-inside-close-paren-matching open paren.
(defmacro giraffe-setf-matching-paren-indent (column)
	`(save-excursion
		(and (goto-matched-parenthesis)
				 (progn (goto-bol)
					 (skip-chars-forward " \t")
					 (setf ,column (current-column))))))

(defun calc-giraffe-indent ()
	(save-excursion
		(setq column nil)
		;previous line
		(when (giraffe-previous-line)
			(goto-bol)
			(skip-chars-forward " \t")
			(setq column (current-column))
			(setq last-open-paren -1)
			(setq last-close-paren -1)
			(loop
				(cond
				 ((looking-at "[\n]") (return))
				 ((parse-point-syntax) ())
				 ((looking-at "[[({]") (setq last-open-paren (current-column)))
				 ((looking-at "[])}]") (progn (setq last-close-paren (current-column))
																 (giraffe-setf-matching-paren-indent column))))
				 (forward-char))
			(when (< last-close-paren last-open-paren) (incf column *giraffe-indent-level*))))
		;current line
	(save-excursion
		(goto-bol)
		(skip-chars-forward " \t")
		(when (looking-at "[]})]") (giraffe-setf-matching-paren-indent column)))
	column)

(defun giraffe-newline-and-indent (&optional (arg 1))
	(interactive "*p")
	(delete-trailing-spaces)
	(insert #\LFD arg)
	(giraffe-indent-line))

(defun giraffe-end-block ()
	(interactive "*")
	(unless (prog1
						(parse-point-syntax)
						(self-insert-command))
		(giraffe-indent-line))
	(save-excursion
		(backward-char)
		(and (goto-matched-parenthesis)
				 (show-matched-parenthesis))))
;------end indent functions---------------------

(defun giraffe-comment-region (from to)
	(interactive "*r")
	(if (> from to)
			(rotatef from to))
	(save-excursion
		(save-restriction
			(narrow-to-region (point-min) to)
			(goto-char from)
			(replace-buffer "^" "#" :regexp t))))

(defun giraffe-uncomment-region (from to)
	(interactive "*r")
	(if (> from to)
			(rotatef from to))
	(save-excursion
		(save-restriction
			(narrow-to-region (point-min) to)
			(goto-char from)
			(replace-buffer "^#" "" :regexp t))))


(defun giraffe-mode ()
	(interactive)
	(kill-all-local-variables)
	(setq mode-name "Giraffe")
	(setq buffer-mode 'giraffe-mode)
	(use-syntax-table *giraffe-mode-syntax-table*)
	(use-keymap *giraffe-mode-map*)
	(make-local-variable 'mode-specific-indent-command)
	(setq mode-specific-indent-command 'giraffe-indent-line)
	(make-local-variable 'indent-tabs-mode)
	(setq indent-tabs-mode *giraffe-indent-tabs-mode*)
	(and *giraffe-keyword-file*
			 (null *giraffe-keyword-hash-table*)
			 (setq *giraffe-keyword-hash-table*
						 (load-keyword-file *giraffe-keyword-file*)))
	(when *giraffe-keyword-hash-table*
		(make-local-variable 'keyword-hash-table)
		(setq keyword-hash-table *giraffe-keyword-hash-table*))
	(run-hooks '*giraffe-mode-hook*))
